/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

int verb_show_environment(int argc, char *argv[], void *userdata);
int verb_set_environment(int argc, char *argv[], void *userdata);
int verb_import_environment(int argc, char *argv[], void *userdata);
